// will be nice to put all the private methods in a closure, but
// putting them all in top scope makes it a lot easier for debugging

var wrappingDom = document.querySelector('#mbx-wrapper');

var editingDom = null;
var replyOrForwardDom = null;
var signatureDom = null;

var replyOrForwardDomWatcher = null;
var signatureDomWatcher = null;

// to be called after editing
// 1) informs compose view that html is updated
// 2) informs compose view that cursor position is updated
function editUpdate() {
    requestUpdateHtml();
    requestUpdateCursor();
}

var updateEvents = ['input', 'keyup', 'paste', 'cut'];
function makeEditWatcher(dom) {
    var edited = false;
    if (dom.classList.contains("edited")) {
        edited = true;
    }

    var markEditFunc;

    if (!edited) {
        markEditFunc = function() {
            updateEvents.forEach(function(e) {
                edited = true;
                dom.classList.add("edited");
                dom.removeEventListener(e, markEditFunc);
            })
        }
        updateEvents.forEach(function(e) {
            dom.addEventListener(e, markEditFunc);
        });
    } else {
        markEditFunc = function() {};
    }

    return {
        isEdited: function() { return edited; },
        markEdited: markEditFunc
    }
}

function makeNotifyHtmlChange(dom) {
    updateEvents.forEach(function(e) {
        dom.addEventListener(e, editUpdate);
    })
}

function makeEditingDom() {
    editingDom = document.querySelector('#mbx-content');

    if (!editingDom) {
        editingDom = document.createElement("div");
        editingDom.id = "mbx-content";
        editingDom.contentEditable = true;
        wrappingDom.insertBefore(editingDom, wrappingDom.firstChild);
    }

    makeNotifyHtmlChange(editingDom);
}

makeEditingDom();

function makeSignatureDom(createIfNotExist) {
    signatureDom = document.querySelector('#mbx-wrapper>.mailbox_signature');

    if (!signatureDom) {
        if (!createIfNotExist) return;
        signatureDom = document.createElement('div');
        signatureDom.className = 'mailbox_signature';
        signatureDom.contentEditable = true;
        wrappingDom.insertBefore(signatureDom, editingDom.nextSibling);
    }

    makeNotifyHtmlChange(signatureDom);
    signatureDomWatcher = makeEditWatcher(signatureDom);
    return signatureDom;
}

function unbreakableNode(node) {
    return node.tagName === "MAILBOX-IMG";
}

//recursive function to move the inline edit that is a child of the passed in parentNode out of its parent and up the chain to the containing div#mbx-old-email
function moveInlineEditFromNode(parentNode, inlineEdit) {

    //if we've reached the containing div then return
    if(parentNode === replyOrForwardDom) return;

    if (unbreakableNode(parentNode)) {
        parentNode.parentNode.insertBefore(inlineEdit, parentNode.nextSibling);
        moveInlineEditFromNode(parentNode.parentNode, inlineEdit);
        return;
    }

    //inlineEdit is the node that we're moving out
    var inlineEdit;

    //clone the current node we're checking
    var cloneNode = parentNode.cloneNode(false);

    //keep track of when to leave the children (before we've found inlineEdit) and when to remove them (after)
    var removeChildren = false;

    //keep track of all the children after inlineEdit that we'll have to remove and add to cloneNode
    var childrenToRemove = [];

    var i;
    for (i = 0; i < parentNode.childNodes.length; i++) {

        //if this child is the inlineEdit then save it, and set remove children to true;
        //if it's not the inlineEdit and removeChildren is true then add it to the array of childrenToRemove
        var child = parentNode.childNodes[i];
        if(child === inlineEdit) {
            removeChildren = true;
        } else {
            if(removeChildren) {
                childrenToRemove.push(child);
            }
        }
    }

    //go through all the children to remove and add them to cloneNode
    var e;
    for (e = 0; e < childrenToRemove.length; e++) {
        var child = childrenToRemove[e];
        child.parentNode.removeChild(child);
        cloneNode.appendChild(child);
    }

    //insert the cloneNode directly after parentNode
    parentNode.parentNode.insertBefore(cloneNode, parentNode.nextSibling);

    //insert the inlineEdit directly in between parentNode and cloneNode
    cloneNode.parentNode.insertBefore(inlineEdit, cloneNode);

    //call with the next parentNode
    moveInlineEditFromNode(parentNode.parentNode, inlineEdit);
}

//find the dummy inlineEdit node, move it out to div#mbx-old-email, and replace it with text to type the inline edit in
function moveInlineEdit(inlineEditNode) {

    //move it out to div#mbx-old-email
    moveInlineEditFromNode(inlineEditNode.parentNode, inlineEditNode);

    //replace dummy inlineEditNode with text div containing empty string
    var textDiv = document.createElement("span");
    textDiv.setAttribute("class", "mailbox-inline-edit");
    textDiv.textContent = "\u200B";
    inlineEditNode.parentNode.replaceChild(textDiv, inlineEditNode);
    var range = document.createRange();
    range.selectNodeContents(textDiv);
    range.collapse(false);
    var selection = window.getSelection();
    selection.removeAllRanges();
    selection.addRange(range);

    // insert a new line so that the next new line won't be ignored
    var br = document.createElement("br");
    range = getSelection().getRangeAt(0);
    range.insertNode(br);
    range.setStartBefore(br);
    range.setEndBefore(br);
    selection = getSelection();
    selection.removeAllRanges();
    selection.addRange(range);
    range.collapse(false);
}

function shouldAllowEnter(sel) {
    var container = sel.getRangeAt(0).startContainer;
    while(!container.tagName && container.parentNode) {
        container = container.parentNode;
    }
    return container.classList && container.classList.contains("mailbox-inline-edit");
}

function makeReplyOrForwardDom(createIfNotExist) {
    replyOrForwardDom = document.querySelector('#mbx-old-email');
    if (!replyOrForwardDom) {
        if (!createIfNotExist) return;
        replyOrForwardDom = document.createElement("div");
        replyOrForwardDom.id = "mbx-old-email";
        // wrappingDom.appendChild(replyOrForwardDom);
        wrappingDom.insertBefore(replyOrForwardDom, null);
    }

    replyOrForwardDom.contentEditable = true;

    // when user press enter, break the block quote so that we have something like:
    // >> quoted stuff...
    // user's comment
    // >> continue qutoed stuff...
    replyOrForwardDom.addEventListener("keyup", function(e) {
        if (e.keyCode == 13) {
            if (shouldAllowEnter(getSelection())) {
                // unfortunately, execCommand(insertHTML) does not work as expected
                // as it will move the cursor to the next element
                var br = document.createElement("br");
                var selection = getSelection();
                var range = selection.getRangeAt(0);
                range.insertNode(br);
                range.setStartAfter(br);
                range.setEndAfter(br);
                selection.removeAllRanges();
                selection.addRange(range);
                range.collapse(false);
            } else {
                e.preventDefault();
                e.stopPropagation();

                var inlineEditNode = document.createElement("span");
                window.getSelection().getRangeAt(0).insertNode(inlineEditNode);
                moveInlineEdit(inlineEditNode);
            }
        }
        editUpdate();
    });

    replyOrForwardDom.addEventListener("input", function(e) {
        editUpdate();
    })

    replyOrForwardDom.addEventListener("keydown", function(e) {
        if (e.keyCode == 13) {
            // to disable default enter key behaviour, we have to do
            // it here in the keydown event
            e.preventDefault();
            e.stopPropagation();
        }
    });

    replyOrForwardDomWatcher = makeEditWatcher(replyOrForwardDom);
    return replyOrForwardDom;
}

// tells the webview that the cursor has moved
function updateCursor() {
    var coor = getCoordOfCursor();
    MbxJavascriptInterface.onCursorMove(
        Math.round(coor.left),
        Math.round(coor.top),
        Math.round(coor.bottom)
    );
}

// tells the webview that the html content has been updated
// and calls back with the updated html
function updateHtml() {
    MbxJavascriptInterface.onHtmlChange(getFormattedHtml());
}

// schedule a call to func. The return function can be called
// multiple times by various events, and it makes
// sure that consecurtive calls will only result
// in one final defer call
function makeDeferCall(func) {
    var timer = null;
    return function() {
        if (timer) {
            clearTimeout(timer);
        }
        timer = setTimeout(func, 10);
    }
}

var requestUpdateCursor = makeDeferCall(updateCursor);
var requestUpdateHtml = makeDeferCall(updateHtml);

var requestUpdateQuotedHtml = makeDeferCall(function() {
    replyOrForwardDom.setAttribute('class', 'edited');
    requestUpdateHtml();
});

function getCoordOfCursor() {
    var sel = window.getSelection();
    var rect = null;

    if (sel.rangeCount) {
        var range = sel.getRangeAt(0);
        if (range.getClientRects) {
            var rects = range.getClientRects();
            if (rects.length > 0) {
                rect = rects[0];
            } else {
                // insert dummy DOM node to get the position,
                var span = document.createElement("span");
                span.textContent = "\u200B";
                range.insertNode(span);
                rect = span.getClientRects()[0];
                span.parentNode.removeChild(span);
            }
        }
    }
    // sometimes top of body becomes negative....
    var baseTop = document.body.getBoundingClientRect().top;
    return {
        top: rect.top - baseTop,
        bottom: rect.bottom - baseTop,
        left: rect.left
    };
}

function cleanup() {
    [].forEach.call(wrappingDom.querySelectorAll("blockquote * .mailbox-inline-edit"), function(elm) {
        elm.className = "oldtagunused";
    });
}

function bytesToSize(bytes) {
    var i, sizes;
    sizes = ['Bytes', 'KB', 'MB', 'GB', 'TB'];
    if (bytes === 0) return '0 bytes';
    i = Math.log(bytes) / Math.log(1024);
    i = i | 0;
    return (bytes / Math.pow(1024, i)).toFixed(1) + ' ' + sizes[i];
}

function shouldSendFullBody() {
    return (signatureDomWatcher && signatureDomWatcher.isEdited()) ||
            (replyOrForwardDomWatcher && replyOrForwardDomWatcher.isEdited());
}

function removeTagIfPresent(root, queryString) {
    var tag = root.querySelector(queryString);
    if (tag) {
        tag.parentNode.removeChild(tag);
    }
}

function getFormattedHtml(forSending) {
    var node = wrappingDom.cloneNode(true);
    if (forSending) {
        if (!shouldSendFullBody()) {
            removeTagIfPresent(node, '#mbx-old-email');
            var toRemove = null;
            // remove top-level signature
            [].forEach.call(node.childNodes, function(elm) {
                if (elm.classList.contains('mailbox_signature')) {
                    toRemove = elm;
                }
            });
            if (toRemove) {
                toRemove.parentNode.removeChild(toRemove);
            }
        }
        [].forEach.call(node.querySelectorAll("#mbx-old-email, #mbx-content, .mailbox_signature"), function(elm) {
            elm.removeAttribute("id");
            elm.removeAttribute("contentEditable");
            elm.removeAttribute("class");
        });
    }
    var images = node.getElementsByClassName("image-attachment");
    [].forEach.call(images, function(image) {
        //if it doesn't have all the required attributes then fail
        var filename = image.getAttribute("data-filename");
        var size = image.getAttribute("data-size");
        var mime = image.getAttribute("data-mime");
        var src = image.getAttribute("data-src");
        if(filename === undefined || size === undefined || mime === undefined || src === undefined) return;

        //delete all the temporary "data-" attributes that become part of the mailboxImg tag
        image.removeAttribute("data-filename");
        image.removeAttribute("data-size");
        image.removeAttribute("data-mime");
        image.removeAttribute("data-src");

        //the javascript DOM doesn't work nicely with creating and modifying custom tags so unfortunately this isn't as easy as creating
        //a mailbox-img tag and setting attributes.  so instead we create a dummy span tag and make its innerHTML the literal HTML of the
        //mailbox-img tag.  then we replace the img tag with the span tag and finally remove the span tag, leaving just the <mailbox-img>
        //tag where the <img> tag was.
        var span = document.createElement("span");
        var html = "<mailbox-img filename=\"" + filename + "\" size=\"" + size + "\" mime=\"" + mime + "\" src=\"" + src + "\"></mailbox-img>";
        span.innerHTML = html;
        var mailboxImg = span.firstChild;
        image.parentNode.replaceChild(span, image);
        mailboxImg.appendChild(image);
        span.parentNode.replaceChild(mailboxImg, span);
    });
    var dbAttachments = node.querySelectorAll(".orc-attach.db-attach");
    [].forEach.call(dbAttachments, function(dbAttachment) {
        var url = dbAttachment.getAttribute("data-url");
        var filename = dbAttachment.getAttribute("data-filename");
        var previewURL = dbAttachment.getAttribute("data-previewurl");
        var size = dbAttachment.getAttribute("data-size");
        var anchorTag = document.createElement("a");
        anchorTag.setAttribute("href", url);
        anchorTag.setAttribute("class", "mb-db-attachment-link");
        anchorTag.setAttribute("data-url", url);
        anchorTag.setAttribute("data-filename", filename);
        anchorTag.setAttribute("data-previewurl", previewURL);
        anchorTag.setAttribute("data-size", size);
        anchorTag.textContent = "Click to view '" + filename + "' on Dropbox";

        var div = document.createElement("div");
        var br = document.createElement("br");
        div.appendChild(anchorTag);
        div.appendChild(br);
        dbAttachment.parentNode.replaceChild(div, dbAttachment);
    });

    var attachments = node.querySelectorAll(".orc-attach.normal-attach");
    [].forEach.call(attachments, function(attachment) {
        var filename = attachment.getAttribute("data-filename");
        var size = attachment.getAttribute("data-size");
        var mime = attachment.getAttribute('data-mime');
        var source = attachment.getAttribute('data-url');
        var filetype_url = attachment.getAttribute('data-filetype_url');

        var span = document.createElement("span");
        var html = "<mailbox-attach filename=\"" + filename +"\" size=\"" + size + "\" mime=\"" + mime + "\" src=\"" + source + "\" filetype_url=\"" + filetype_url + "\"></mailbox-attach>";
        span.innerHTML = html;
        var attachmentTag = span.firstChild;

        attachment.parentNode.replaceChild(attachmentTag, attachment);
    });

    return node.innerHTML;
};

function getAttachmentsSize() {
    var size = 0;
    var attachments = document.querySelectorAll(".orc-attach.normal-attach");
    [].forEach.call(attachments, function(attachment) {
        size += parseInt(attachment.getAttribute('data-size'));
    });
    var images = document.querySelectorAll("img.image-attachment");
    [].forEach.call(images, function(img) {
        size += parseInt(img.getAttribute('data-size'));
    });
    return size;
}

function setupAttachmentDom(dom) {
    var filename = dom.getAttribute("filename");
    var size = parseInt(dom.getAttribute("size"));
    var mime = dom.getAttribute('mime');
    var source = dom.getAttribute('src');

    var div = createAttachmentDom(filename, size, mime, source);

    dom.parentNode.removeChild(dom);
    wrappingDom.appendChild(div);
}

function setupDropboxAttachmentDom(dom) {
    var url = dom.getAttribute("href");
    var filename = dom.getAttribute("data-filename");
    var previewURL = dom.getAttribute("data-previewurl");
    var size = dom.getAttribute("data-size");

    var div = document.createElement("div");

    div.setAttribute("data-url", url);
    div.setAttribute("data-filename", filename);
    div.setAttribute("data-previewurl", previewURL);
    div.setAttribute("data-size", size);

    div.className = "orc-attach db-attach";

    var fileExtension = url.substring(url.lastIndexOf("."));
    var filetype_url = genericIconURL;
    if(fileExtensionMapping.hasOwnProperty(fileExtension)) {
        filetype_url = fileExtensionMapping[fileExtension];
    }
    var sizeText = bytesToSize(size);
    div.innerHTML = "<img src=\"" + filetype_url + "\" class=\"orc-attach-icon\"/><span class=\"orc-attach-name\">" + filename + "</span><br><span class=\"orc-attach-download-info\">" + sizeText + "</span>";
    div.contentEditable = false;

    dom.parentNode.replaceChild(div, dom);
}

function setupImageAttachmentDom(dom) {
    var filename = dom.getAttribute("filename");
    var size = dom.getAttribute("size");
    var mime = dom.getAttribute("mime");
    var src = dom.getAttribute("src");

    var img = document.createElement('img');
    img.className = "image-attachment";
    img.setAttribute('data-filename', filename);
    img.setAttribute('data-size', size);
    img.setAttribute('data-mime', mime);
    img.setAttribute('data-src', src);

    dom.parentNode.replaceChild(img, dom);

    MbxJavascriptInterface.loadImage(src, mime);
}

function createAttachmentDom(filename, size, mime, source) {
    var div = document.createElement('div');
    div.className = "orc-attach normal-attach";
    div.setAttribute("data-filename", filename);
    div.setAttribute("data-size", size);
    div.setAttribute("data-mime", mime);
    div.setAttribute("data-url", source);

    var sizeText = bytesToSize(size);
    var filetype_url = genericIconURL;
    if(mimeTypeMapping.hasOwnProperty(mime)) {
        filetype_url = mimeTypeMapping[mime];
    }

    var inner = document.createElement('div');
    inner.className = "attachment-inner";
    inner.addEventListener("click", function() {
        MbxJavascriptInterface.onAttachmentClicked(filename, source, size, mime);
    });

    var remove = document.createElement('div');
    remove.className = "remove";
    remove.innerHTML = "<img src=\"" + attachmentDeleteIconURL + "\">"

    remove.addEventListener('click', function() {
        div.parentNode.removeChild(div);
        editUpdate();
    });

    inner.innerHTML = "<img src=\"" + filetype_url + "\" class=\"orc-attach-icon\"/><span class=\"orc-attach-name\">" + filename + "</span><br><span class=\"orc-attach-download-info\">" + sizeText + "</span>";
    div.appendChild(inner);
    div.appendChild(remove);
    return div;
}

function setupAttachments() {
    [].forEach.call(document.querySelectorAll('mailbox-img'), setupImageAttachmentDom);
    [].forEach.call(document.querySelectorAll('mailbox-attach'), setupAttachmentDom);
    [].forEach.call(document.querySelectorAll('.mb-db-attachment-link'), setupDropboxAttachmentDom);
}

// troublesome dom refers to elements that sometimes cannot be deleted correctly by backspace
// in some versions of android
function isTroublesomeDom(dom) {
    if (!dom) return false;
    return dom.tagName === 'IMG' || dom.contentEditable === "false";
}

function getLastTroublesomeDescendant(node) {
    if (isTroublesomeDom(node)) {
        return node;
    }

    if (node.childNodes) {
        return getLastTroublesomeDescendant(node.lastChild);
    }

    return null;
}

function getTroublesomePreviousDom(dom, offset) {
    if ([editingDom, signatureDom, replyOrForwardDom].indexOf(dom) !== -1) {
        return null;
    }

    if (offset) {
        if (dom.tagName) {
            return getLastTroublesomeDescendant(dom.childNodes[offset - 1]);
        } else {
            return null;
        }
    }

    if (dom.previousSibling) {
        return getLastTroublesomeDescendant(dom.previousSibling);
    } else {
        return getTroublesomePreviousDom(dom.parentNode);
    }
}

function requestLayout() {
    MbxJavascriptInterface.requestLayout();
}

// execCommand('InsertHTML',,) can fail if we don't have a cursor in the webview
// which can happen if e.g. the user has not tapped on the webview yet
// in this case, we just append the things we want to insert to the end
var dummyContainer = document.createElement('div');
function insertHTML(html) {
    if (getSelection().rangeCount) {
        document.execCommand("InsertHTML", false, html);
    } else {
        // we don't currently have a cursor, insert at the end
        dummyContainer.innerHTML = html;
        var fragment = document.createDocumentFragment();
        [].forEach.call(dummyContainer.childNodes, function(node) {
            fragment.appendChild(node.cloneNode(true));
        });
        dummyContainer.innerHTML = "";
        editingDom.appendChild(fragment);

        setTimeout(function() {
            // now move the cursor be after the inserted elements
            // delay because we may not have focus yet
            var range = document.createRange();
            var sel = window.getSelection();
            range.setStart(editingDom, editingDom.childNodes.length);
            range.collapse(true);
            sel.removeAllRanges();
            sel.addRange(range);
            editUpdate();
        }, 50)
    }
}

// methods to be called by the webview should be here.
var editable = {
    setHtml: function(html) {
        var wrap = document.createElement('div');
        wrap.innerHTML = html;
        if (wrap.querySelector('#mbx-wrapper')) {
            wrappingDom.innerHTML = wrap.querySelector('#mbx-wrapper').innerHTML;
        } else if (wrap.querySelector('#mbx-content')) {
            wrappingDom.innerHTML = html;
            makeEditingDom();
        } else {
            makeEditingDom();
            editingDom.innerHTML = html;
        }
        makeSignatureDom();
        makeReplyOrForwardDom();
        cleanup();
        setupAttachments();
        requestUpdateHtml();
    },
    updateCursor: updateCursor,
    setReplyOrForwardContent: function(htmlContent) {
        var replyOrForwardDom = makeReplyOrForwardDom(true);
        replyOrForwardDom.innerHTML = htmlContent;
        cleanup();
        requestUpdateHtml();
    },
    gainFocus: function() {
        setTimeout(function() {
            if (document.activeElement !== document.body && document.activeElement) {
                document.activeElement.focus();
            } else {
                editingDom.focus();
            }
            requestUpdateCursor();
        }, 0);
    },
    insertOrUpdateSignature: function(htmlContent) {
        if (signatureDomWatcher && signatureDomWatcher.isEdited()) return;
        var signatureDom = makeSignatureDom(true);
        signatureDom.innerHTML = "<br>" + htmlContent;
        signatureDom.style.display = 'block';
        requestUpdateHtml();
    },
    prepareToSendEmail: function() {
        var html = getFormattedHtml(true);
        var attachmentSize = getAttachmentsSize();
        var sendFullbody = shouldSendFullBody();
        MbxJavascriptInterface.onEmailPrepared(html, attachmentSize, sendFullbody);
    },
    insertImage: function(attId, src, mimeType, filename, size) {
        var imagePath = MbxJavascriptInterface.getImageSrc(src);
        editable.gainFocus();
        var html = "<br><img id=\"" + attId + "\" class=\"image-attachment\" src=\"" + imagePath + "\" data-filename=\"" + filename + "\" data-size=\"" + size + "\" data-mime=\"" + mimeType + "\" data-src=\"" + src + "\"/><div><br></div>";
        insertHTML(html);
    },
    insertDropboxAttachment: function(preview_url, file_name, url, size) {
        editable.gainFocus();
        var fileExtension = url.substring(url.lastIndexOf("."));
        var filetype_url = genericIconURL;
        if(fileExtensionMapping.hasOwnProperty(fileExtension)) {
            filetype_url = fileExtensionMapping[fileExtension];
        }
        var sizeText = bytesToSize(size);
        var html = "<div><br></div><div class=\"orc-attach db-attach\" data-previewurl=\"" + preview_url + "\" data-filename=\"" + file_name + "\" data-url=\"" + url + "\" data-size=\"" + size + "\"><img src=\"" + filetype_url + "\" class=\"orc-attach-icon\"/><span class=\"orc-attach-name\">" + file_name + "</span><br><span class=\"orc-attach-download-info\">" + sizeText + "</span></div><div><br></div>";
        insertHTML(html);
        [].forEach.call(document.querySelectorAll(".orc-attach"), function(elm) {
            elm.contentEditable = false;
        });
        editUpdate();
    },
    insertAttachment: function(attId, mime_type, file_name, file_path, url, size) {
        var filetype_url = genericIconURL;
        if(mimeTypeMapping.hasOwnProperty(mime_type)) {
            filetype_url = mimeTypeMapping[mime_type];
        }
        var sizeText = bytesToSize(size);

        var node = createAttachmentDom(file_name, size, mime_type, url);

        wrappingDom.appendChild(node);
        editUpdate();
    },
    maybeHelpDelete: function() {
        var range = getSelection().getRangeAt(0);
        var previous = getTroublesomePreviousDom(range.startContainer, range.startOffset);
        if (previous) {
            setTimeout(function() {
                var parent = previous.parentNode;
                if (parent) {
                    parent.removeChild(previous);
                }
            }, 0);
        }
    },
    loadedImage: function(src) {
        var imgSrc = MbxJavascriptInterface.getImageSrc(src);
        [].forEach.call(document.querySelectorAll("img[data-src=\"" + src + "\"]"), function(img) {
            img.src = imgSrc;
        });
        setTimeout(requestLayout, 100);
    }
};

// to prevent RigidWebView from dropping the height change
MbxJavascriptInterface.ready();
